@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2017 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtBluetooth module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2022 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qbluetoothdevicediscoveryagent.h"
 #include "qbluetoothdevicediscoveryagent_p.h"
@@ -50,7 +14,6 @@
 #include <QtCore/QMutex>
 #include <QtCore/qendian.h>
 
-#include <winrt/Windows.Devices.Radios.h>
 #include <winrt/Windows.Devices.Bluetooth.h>
 #include <winrt/Windows.Devices.Bluetooth.Advertisement.h>
 #include <winrt/Windows.Devices.Bluetooth.Rfcomm.h>
@@ -70,6 +33,9 @@ using namespace winrt::Windows::Storage::Streams;
 
 QT_BEGIN_NAMESPACE
 
+QT_IMPL_METATYPE_EXTERN(ManufacturerData)
+QT_IMPL_METATYPE_EXTERN(ServiceData)
+
 Q_DECLARE_LOGGING_CATEGORY(QT_BT_WINDOWS)
 
 // Endianness conversion for quint128 doesn't exist in qtendian.h
@@ -102,6 +68,37 @@ static ManufacturerData extractManufacturerData(const BluetoothLEAdvertisement &
     return ret;
 }
 
+static ServiceData extractServiceData(const BluetoothLEAdvertisement &ad)
+{
+    static constexpr int serviceDataTypes[3] = { 0x16, 0x20, 0x21 };
+
+    ServiceData ret;
+
+    for (const auto &serviceDataType : serviceDataTypes) {
+        const auto dataSections = ad.GetSectionsByType(serviceDataType);
+        for (const auto &section : dataSections) {
+            const unsigned char dataType = section.DataType();
+            const QByteArray bufferData = byteArrayFromBuffer(section.Data());
+            if (dataType == 0x16) {
+                Q_ASSERT(bufferData.size() >= 2);
+                ret.insert(QBluetoothUuid(qFromLittleEndian<quint16>(bufferData.constData())),
+                           bufferData + 2);
+            } else if (dataType == 0x20) {
+                Q_ASSERT(bufferData.size() >= 4);
+                ret.insert(QBluetoothUuid(qFromLittleEndian<quint32>(bufferData.constData())),
+                           bufferData + 4);
+            } else if (dataType == 0x21) {
+                Q_ASSERT(bufferData.size() >= 16);
+                ret.insert(QBluetoothUuid(qToBigEndian<quint128>(
+                                   qFromLittleEndian<quint128>(bufferData.constData()))),
+                           bufferData + 16);
+            }
+        }
+    }
+
+    return ret;
+}
+
 // Needed because there is no explicit conversion
 static GUID fromWinRtGuid(const winrt::guid &guid)
 {
@@ -127,6 +124,7 @@ public:
     }
     void init()
     {
+        m_watcher.AllowExtendedAdvertisements(true);
         m_watcher.ScanningMode(BluetoothLEScanningMode::Active);
     }
     void start() {
@@ -146,6 +144,7 @@ signals:
     // so we need to use Qt::QueuedConnection
     void advertisementDataReceived(quint64 address, qint16 rssi,
                                    const ManufacturerData &manufacturerData,
+                                   const ServiceData &serviceData,
                                    const QList<QBluetoothUuid> &uuids);
 private:
     void subscribeToEvents()
@@ -161,6 +160,7 @@ private:
                 const BluetoothLEAdvertisement ad = args.Advertisement();
 
                 const ManufacturerData manufacturerData = extractManufacturerData(ad);
+                const ServiceData serviceData = extractServiceData(ad);
 
                 QList<QBluetoothUuid> serviceUuids;
                 const auto guids = ad.ServiceUuids();
@@ -170,7 +170,7 @@ private:
                 }
 
                 emit thisPtr->advertisementDataReceived(address, rssi, manufacturerData,
-                                                        serviceUuids);
+                                                        serviceData, serviceUuids);
             });
     }
     void unsubscribeFromEvents()
@@ -240,7 +240,7 @@ private:
 Q_SIGNALS:
     void deviceFound(const QBluetoothDeviceInfo &info);
     void deviceDataChanged(const QBluetoothAddress &address, QBluetoothDeviceInfo::Fields,
-                           qint16 rssi, ManufacturerData manufacturerData);
+                           qint16 rssi, ManufacturerData manufacturerData, ServiceData serviceData);
     void errorOccured(QBluetoothDeviceDiscoveryAgent::Error error);
     void scanFinished();
 
@@ -250,6 +250,7 @@ private slots:
 
     void onAdvertisementDataReceived(quint64 address, qint16 rssi,
                                      const ManufacturerData &manufacturerData,
+                                     const ServiceData &serviceData,
                                      const QList<QBluetoothUuid> &uuids);
 
     void stopAdvertisementWatcher();
@@ -258,6 +259,7 @@ private:
     struct LEAdvertisingInfo {
         QList<QBluetoothUuid> services;
         ManufacturerData manufacturerData;
+        ServiceData serviceData;
         qint16 rssi = 0;
     };
 
@@ -402,7 +404,7 @@ void QWinRTBluetoothDeviceDiscoveryWorker::onDeviceEnumerationCompleted(int watc
 // this function executes in main worker thread
 void QWinRTBluetoothDeviceDiscoveryWorker::onAdvertisementDataReceived(
         quint64 address, qint16 rssi, const ManufacturerData &manufacturerData,
-        const QList<QBluetoothUuid> &uuids)
+        const ServiceData &serviceData, const QList<QBluetoothUuid> &uuids)
 {
     // Merge newly found services with list of currently found ones
     bool needDiscoverServices = false;
@@ -421,7 +423,12 @@ void QWinRTBluetoothDeviceDiscoveryWorker::onAdvertisementDataReceived(
                 if (adInfo.manufacturerData != m_foundLEDevicesMap[address].manufacturerData)
                     changedFields.setFlag(QBluetoothDeviceInfo::Field::ManufacturerData);
             }
-            for (const QBluetoothUuid &uuid : qAsConst(uuids)) {
+            if (adInfo.serviceData != serviceData) {
+                m_foundLEDevicesMap[address].serviceData.insert(serviceData);
+                if (adInfo.serviceData != m_foundLEDevicesMap[address].serviceData)
+                    changedFields.setFlag((QBluetoothDeviceInfo::Field::ServiceData));
+            }
+            for (const QBluetoothUuid &uuid : std::as_const(uuids)) {
                 if (!foundServices.contains(uuid)) {
                     foundServices.append(uuid);
                     needDiscoverServices = true;
@@ -433,7 +440,8 @@ void QWinRTBluetoothDeviceDiscoveryWorker::onAdvertisementDataReceived(
                                               Q_ARG(QBluetoothAddress, QBluetoothAddress(address)),
                                               Q_ARG(QBluetoothDeviceInfo::Fields, changedFields),
                                               Q_ARG(qint16, rssi),
-                                              Q_ARG(ManufacturerData, manufacturerData));
+                                              Q_ARG(ManufacturerData, manufacturerData),
+                                              Q_ARG(ServiceData, serviceData));
                 }
             }
             m_foundLEDevicesMap[address].services = foundServices;
@@ -442,6 +450,7 @@ void QWinRTBluetoothDeviceDiscoveryWorker::onAdvertisementDataReceived(
             LEAdvertisingInfo info;
             info.services = std::move(uuids);
             info.manufacturerData = std::move(manufacturerData);
+            info.serviceData = std::move(serviceData);
             info.rssi = rssi;
             m_foundLEDevicesMap.insert(address, info);
         }
@@ -502,7 +511,8 @@ void QWinRTBluetoothDeviceDiscoveryWorker::invokeDeviceFoundWithDebug(const QBlu
 {
     qCDebug(QT_BT_WINDOWS) << "Discovered BTLE device: " << info.address() << info.name()
                            << "Num UUIDs" << info.serviceUuids().size() << "RSSI:" << info.rssi()
-                           << "Num manufacturer data" << info.manufacturerData().size();
+                           << "Num manufacturer data" << info.manufacturerData().size()
+                           << "Num service data" << info.serviceData().size();
 
     QMetaObject::invokeMethod(this, "deviceFound", Qt::AutoConnection,
                               Q_ARG(QBluetoothDeviceInfo, info));
@@ -641,6 +651,7 @@ void QWinRTBluetoothDeviceDiscoveryWorker::handleLowEnergyDevice(const Bluetooth
     const LEAdvertisingInfo adInfo = m_foundLEDevicesMap.value(address);
     m_leDevicesMutex.unlock();
     const ManufacturerData manufacturerData = adInfo.manufacturerData;
+    const ServiceData serviceData = adInfo.serviceData;
     const qint16 rssi = adInfo.rssi;
 
     QBluetoothDeviceInfo info(QBluetoothAddress(address), btName, 0);
@@ -648,6 +659,8 @@ void QWinRTBluetoothDeviceDiscoveryWorker::handleLowEnergyDevice(const Bluetooth
     info.setRssi(rssi);
     for (quint16 key : manufacturerData.keys())
         info.setManufacturerData(key, manufacturerData.value(key));
+    for (QBluetoothUuid key : serviceData.keys())
+        info.setServiceData(key, serviceData.value(key));
     info.setCached(true);
 
     // Use the services obtained from the advertisement data if the device is not paired
@@ -713,11 +726,9 @@ QWinRTBluetoothDeviceDiscoveryWorker::createAdvertisementWatcher()
 }
 
 QBluetoothDeviceDiscoveryAgentPrivate::QBluetoothDeviceDiscoveryAgentPrivate(
-                const QBluetoothAddress &deviceAdapter,
-                QBluetoothDeviceDiscoveryAgent *parent)
-    :   q_ptr(parent)
+        const QBluetoothAddress &deviceAdapter, QBluetoothDeviceDiscoveryAgent *parent)
+    : q_ptr(parent), adapterAddress(deviceAdapter)
 {
-    Q_UNUSED(deviceAdapter);
     mainThreadCoInit(this);
 }
 
@@ -739,30 +750,19 @@ QBluetoothDeviceDiscoveryAgent::DiscoveryMethods QBluetoothDeviceDiscoveryAgent:
 
 void QBluetoothDeviceDiscoveryAgentPrivate::start(QBluetoothDeviceDiscoveryAgent::DiscoveryMethods methods)
 {
-    // Check the availability and power state of the bluetooth adapter.
-    // This is a check that is specific to Qt 6.2. In later versions we use
-    // QBluetoothLocalDevice which got Windows support in Qt 6.3. Note: here we only
-    // check the default adapter as specifying the adapter address is not supported in Qt 6.2
-    // See QTBUG-106874
-    BluetoothAdapter adapter{nullptr};
-    bool result = await(BluetoothAdapter::GetDefaultAsync(), adapter);
-    if (result && adapter) {
-        using namespace winrt::Windows::Devices::Radios;
-        Radio radio(nullptr);
-        result = await(adapter.GetRadioAsync(), radio);
-        if (!result || !radio || (radio.State() != RadioState::On)) {
-            qCWarning(QT_BT_WINDOWS) << "Bluetooth adapter powered off";
-            lastError = QBluetoothDeviceDiscoveryAgent::PoweredOffError;
-            errorString = QBluetoothDeviceDiscoveryAgent::tr("Device is powered off");
-            emit q_ptr->errorOccurred(lastError);
-            return;
-        }
-    } else {
+    QBluetoothLocalDevice adapter(adapterAddress);
+    if (!adapter.isValid()) {
         qCWarning(QT_BT_WINDOWS) << "Cannot find Bluetooth adapter for device search";
         lastError = QBluetoothDeviceDiscoveryAgent::InvalidBluetoothAdapterError;
         errorString = QBluetoothDeviceDiscoveryAgent::tr("Cannot find valid Bluetooth adapter.");
         emit q_ptr->errorOccurred(lastError);
         return;
+    } else if (adapter.hostMode() == QBluetoothLocalDevice::HostPoweredOff) {
+        qCWarning(QT_BT_WINDOWS) << "Bluetooth adapter powered off";
+        lastError = QBluetoothDeviceDiscoveryAgent::PoweredOffError;
+        errorString = QBluetoothDeviceDiscoveryAgent::tr("Bluetooth adapter powered off.");
+        emit q_ptr->errorOccurred(lastError);
+        return;
     }
 
     if (worker)
@@ -806,7 +806,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::registerDevice(const QBluetoothDevic
             QList<QBluetoothUuid> uuids = iter->serviceUuids();
             uuids.append(info.serviceUuids());
             const QSet<QBluetoothUuid> uuidSet(uuids.begin(), uuids.end());
-            if (iter->serviceUuids().count() != uuidSet.count())
+            if (iter->serviceUuids().size() != uuidSet.size())
                 iter->setServiceUuids(uuidSet.values().toVector());
             if (iter->coreConfigurations() != info.coreConfigurations())
                 iter->setCoreConfigurations(QBluetoothDeviceInfo::BaseRateAndLowEnergyCoreConfiguration);
@@ -821,7 +821,8 @@ void QBluetoothDeviceDiscoveryAgentPrivate::registerDevice(const QBluetoothDevic
 void QBluetoothDeviceDiscoveryAgentPrivate::updateDeviceData(const QBluetoothAddress &address,
                                                              QBluetoothDeviceInfo::Fields fields,
                                                              qint16 rssi,
-                                                             ManufacturerData manufacturerData)
+                                                             ManufacturerData manufacturerData,
+                                                             ServiceData serviceData)
 {
     if (fields.testFlag(QBluetoothDeviceInfo::Field::None))
         return;
@@ -836,6 +837,9 @@ void QBluetoothDeviceDiscoveryAgentPrivate::updateDeviceData(const QBluetoothAdd
             if (fields.testFlag(QBluetoothDeviceInfo::Field::ManufacturerData))
                 for (quint16 key : manufacturerData.keys())
                     iter->setManufacturerData(key, manufacturerData.value(key));
+            if (fields.testFlag(QBluetoothDeviceInfo::Field::ServiceData))
+                for (QBluetoothUuid key : serviceData.keys())
+                    iter->setServiceData(key, serviceData.value(key));
             emit q->deviceUpdated(*iter, fields);
             return;
         }
